/***********************************************************************
 * neopixel.cpp:
 * Controlling Neopixel RGB LED strip by Raspberry Pi
 *
 * File writed by Ferenc Takács (taxi).
 *   Program uses the wiringPi library wich modified by me.
 *   	git clone git://git.drogon.net/wiringPi 
 * 		Please see   http://wiringpi.com/ 
 *      After download the library source
 *      change the wiringPi.h and wiringPi.c files to my versions,
 *      and intall by ./build command.
 *
 * Before run the program, set the LED_NUM, and LED_PIN constant to you need!
 * Check the correct timing (29.4 us per LED)!
 * Tested on:
 *     Pi 2
 *     Pi 3
 * Copyright (c) 2016 Ferenc Takács. <ferenc.takacs.hu@freemail.hu>
 ************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <memory.h>
#include <string.h>
#include <time.h>
#include <sys/time.h>
#include <pthread.h>
#include <wiringPi.h>
#include <unistd.h>
#include <termios.h> 
#include <fcntl.h>
#include <sys/resource.h> 
#include "gamma.h"

template <class T> inline T max(T a, T b) { return a > b ? a: b ; }

typedef volatile unsigned int vuInt; // volatile content
typedef vuInt *pvuInt; // pointer to volatile content

class Neopixel {	
	const int calibrate_num;
	static bool firstCase;
	pthread_mutex_t cs_mutex;
	vuInt dummy_;
	pvuInt dummy;
	// need 29.4 us per led
	unsigned int time_limit;
	// 2*8 timer per byte (2 timers to one bit)
	int timer_num;
	unsigned short *timer;
	bool simulate;
	int led_pin;
	int led_num;
	// 3 byte (RGB) per LED
	int byte_num;
	unsigned short H0,L0,H1,L1; // limits of timer counters
	unsigned int last_time;
	int	priority;
	// private methods
	void timers_fill();
	void before(); // wait the previous pause and lock
	unsigned int after(); // unlock		
public:
	// Set/clear before end, if need. Default is true.
	bool clear_at_end;
	int repeat;
	unsigned char *rgb;
	// public methods
	inline int LedPin() { return led_pin; }
	inline int LedNum() { return led_num; }
	inline int ByteNum() { return byte_num; }
	Neopixel( int led_num, int led_pin, bool calibrate = true );
	~Neopixel();
	void rgb2grb();	
	void calibrating();
	unsigned int send();
	unsigned int clear();
};

bool Neopixel::firstCase = true;

void Neopixel::timers_fill()
{
	int num,bit;
	unsigned char *col;
	unsigned short *ptr;
	for (num = 0, ptr = timer, col = rgb; num < byte_num; num++) {
		unsigned char ch = *col++;
		for(bit=0;bit<8; bit++) {
			if( (ch&(0x80>>bit)) == 0 ) { // bit zero
				*ptr++ = H0;
				*ptr++ = L0;
			}
			else { // bit one
				*ptr++ = H1;
				*ptr++ = L1;
			}
		}
	}
}

void Neopixel::before() // wait the previous pause and lock
{
	unsigned int remaind = micros()-last_time;
	if( remaind < 80 )
		usleep(80-remaind);
	priority = getpriority(PRIO_PROCESS, 0);
	setpriority(PRIO_PROCESS, 0, -20);
	pthread_mutex_lock( &cs_mutex );
	piHiPri(99);
	last_time = micros();		
}

unsigned int Neopixel::after() // unlock
{
	unsigned int now = micros();
	unsigned int microsec = now - last_time;
	piHiPri(0);
	pthread_mutex_unlock( &cs_mutex );
	setpriority(PRIO_PROCESS, 0, priority);
	last_time = now;	
	return microsec;		
}

Neopixel::Neopixel( int led_num, int led_pin, bool calibrate ) :
	calibrate_num(10),
	dummy_(0),
	dummy(&dummy_),
	time_limit((led_num*295)/10+3),
	timer_num(48*led_num),
	timer(new unsigned short[48*max(led_num,calibrate_num)]),
	simulate(false),
	led_pin(led_pin),
	led_num(led_num),
	byte_num(3*led_num),
	H0(89),L0(259),H1(186),L1(163), // default for Pi 3 and Pi 2
	last_time(0),
	clear_at_end(true),
	repeat(0),
	rgb(new unsigned char[3*max(led_num,calibrate_num)])
{
	pthread_mutex_init(&cs_mutex, NULL);
	if( firstCase )
		wiringPiSetup();
	int model,rev,mem,maker,overVolted;
	piBoardId(&model,&rev,&mem,&maker,&overVolted);
	printf("Computer model: %s, rev: %d, mem: %d kB, maker: %s, overVolted: %d\n",
		piModelNames[model],rev,piMemorySize[mem],piMakerNames[maker],overVolted);
	pinMode(led_pin, OUTPUT);
	if( calibrate ) {
		calibrating();			
	}
	firstCase = false;
}

Neopixel::~Neopixel()
{
	if( clear_at_end )
		clear();
	before(); // wait last
	after(); // unlock
	pinMode(led_pin, INPUT);
	pthread_mutex_destroy(&cs_mutex);
	delete []rgb;
	delete []timer;
}

void Neopixel::rgb2grb() // set the correct GRB order to neopixel from standard RGB order
{ 
	for(int i=0; i<byte_num; i+=3){
		char r,g,b;
		r = rgb[i+0];
		g = rgb[i+1];
		b = rgb[i+2];
		rgb[i+0] = g;
		rgb[i+1] = r;
		rgb[i+2] = b;
	}
}

void Neopixel::calibrating()
{
	simulate = true;
	int led_num_s = led_num;
	int byte_num_s = byte_num;
	int timer_num_s = timer_num;
	led_num = calibrate_num;
	byte_num = 3*calibrate_num;
	timer_num = 48*calibrate_num;
	time_limit = 0; // max repeat in send at search the minimum
	H0 = L0 = H1 = L1 = 0;
	double zero_time = send()*0.5; // half the two bits time
	unsigned short const unitNum = 255;
	H0 = L0 = H1 = L1 = unitNum;
	double unit_time = send()*0.5; // half the two bits time
	double range =  unitNum / (unit_time - zero_time);
	// WS2812 bit time: 1250 ns = 350 ns (H0) + 800 ns (L0); 1300 ns = 700 ns (H1) + 600 ns (L1)
	// (+-150ns) pause 50 us
	//    min time:  850 = 200 + 650 (H); 1000 = 550 + 450 (L)
	//    max time: 1450 = 500 + 950 (H); 1600 = 850 + 750 (L)
	// SK6812 bit time: 1200 ns = 300 ns (H0) + 900 ns (L0) = 600 ns (H1) + 600 ns (L1)
	// (+-150ns) pause 80 us
	//    min time:  900 = 150 + 750 (H);   900 = 450 + 450 (L)
	//    max time: 1500 = 450 + 1050 (H); 1500 = 750 + 750 (L)
	// times avarages of WS2812 and SK6812
	// one bit: 1225 ns = 325 ns (H0) + 900 ns (L0) = 650 ns (H1) + 575 ns (L1)
	// one led(24 bits): 29.4 us = 7.8 us (H0) + 21.6 us (L0) = 15.6 us (H1) + 13.8 us (L1)
	H0 = (unsigned short)( ( 7.8*calibrate_num - zero_time) * range);
	L0 = (unsigned short)( (21.6*calibrate_num - zero_time) * range);
	H1 = (unsigned short)( (15.6*calibrate_num - zero_time) * range);
	L1 = (unsigned short)( (13.8*calibrate_num - zero_time) * range);
	unsigned int new_time = send(); // result of calibrate
	printf("Calibrate on %d leds, time %d us, Counters(H0 %d, L0 %d, H1 %d, L1 %d)\n", calibrate_num, new_time, H0, L0, H1, L1);
	
	led_num = led_num_s;
	byte_num = byte_num_s;
	timer_num = timer_num_s;
	time_limit = ((led_num*(new_time+2))/calibrate_num+3);
	simulate = false;
}

unsigned int Neopixel::send()
{
	unsigned int mask;
	pvuInt clr, set;
	timers_fill();
	if( simulate ) {
		clr = set = dummy;
		mask = 1;
	}
	else
		digitalAddress( led_pin, &clr, &set, 0, &mask); // get control addresses, and value
	if( !mask ) return 0; // wrong mode		
	unsigned int microsec=0;
	for(int k=0; k<4; k++) { // must repeats the sending, because thread may interrupt
		before();
		unsigned short *ptr;
		int j;
		for (j=0, ptr = timer; j<timer_num; j++) {
			if( (j&1)!=0 )
				*clr = mask; // control to LOW
			else
				*set = mask; // control to HIGH
			for(int i=*ptr++; i>0; i--) { // timing
				*dummy = i;
			}
		}
		unsigned int microsec2 = after();
		if( microsec > microsec2 || !k )
			microsec = microsec2;
		if( microsec < time_limit )
			break; // not interrupted, send was correct
		if( !k )
			repeat++;
	}
	return microsec;
}

unsigned int Neopixel::clear()
{
	memset(rgb,0,byte_num);
	return send();	
}


class getcharNoWait {
	struct termios oldt;
	int oldf;
public:
	getcharNoWait() // Do not use as auto variable! Use new getcharNoWait()! Do not ask why!
	{
		struct termios newt;
		tcgetattr(STDIN_FILENO, &oldt);
		newt = oldt;
		newt.c_lflag &= ~(ICANON | ECHO);
		tcsetattr(STDIN_FILENO, TCSANOW, &newt);
		oldf = fcntl(STDIN_FILENO, F_GETFL, 0);
		fcntl(STDIN_FILENO, F_SETFL, oldf | O_NONBLOCK);
	}
	~getcharNoWait()
	{
		tcsetattr(STDIN_FILENO, TCSANOW, &oldt);
		fcntl(STDIN_FILENO, F_SETFL, oldf);
	}
};

	// pin_num  0 - GPIO17 (header #11), GPIO_GEN0
	// pin_num  1 - GPIO18 (header #12), GPIO_GEN1, PCM_CLK
	// pin_num  2 - GPIO21 (header #13), GPIO_GEN2, GPIO27 at Rev 2
	// pin_num  3 - GPIO22 (header #15), GPIO_GEN3
	// pin_num  4 - GPIO23 (header #16), GPIO_GEN4
	// pin_num  5 - GPIO24 (header #18), GPIO_GEN5
	// pin_num  6 - GPIO25 (header #22), GPIO_GEN6
	// pin_num  7 - GPIO04 (header #07), GPIO_GCLK
	// pin_num  8 - GPIO00 (header #03), SDA1 I2C, GPIO02 at Rev 2
	// pin_num  9 - GPIO01 (header #05), SCL1 I2C, GPIO03 at Rev 2
	// pin_num 10 - GPIO08 (header #24), SPI0_CE0_N
	// pin_num 11 - GPIO07 (header #26), SPI0_CE1_N
	// pin_num 12 - GPIO10 (header #19), SPI0_MOSI
	// pin_num 13 - GPIO09 (header #21), SPI0_MOSO
	// pin_num 14 - GPIO11 (header #23), SPI0_CLK
	// pin_num 15 - GPIO14 (header #08), UART0_TDX
	// pin_num 16 - GPIO15 (header #10), UART0_RDX
	// pin_num 17 - GPIO28 (header #24), trough GPIO08
	// pin_num 18 - GPIO29 (header #21), trough GPIO09
	// pin_num 19 - GPIO30 (header #19), trough GPIO10
	// pin_num 20 - GPIO31 (header #23), trough GPIO11
	// pin_num 21 - GPIO05 (header #29)
	// pin_num 22 - GPIO06 (header #31)
	// pin_num 23 - GPIO13 (header #33)
	// pin_num 24 - GPIO19 (header #35)
	// pin_num 25 - GPIO26 (header #37)
	// pin_num 26 - GPIO12 (header #32)
	// pin_num 27 - GPIO16 (header #36)
	// pin_num 28 - GPIO20 (header #38)
	// pin_num 29 - GPIO21 (header #40)
int pin_num = 1;

int led_num = 12;

int shift_time = 100;

int color[100] = { // RGB list
	0x000000,	// black, off
	0x0000ff,	// red
	0x00ff00,	// green
  	0xff0000,	// blue
	0x000000,
	0x000000,
	0xffffff,	// white
	0x000000,
	0x000000,
	0xffff00,	// cyan
	0xff00ff,	// magenta
	0x00ffff,	// yellow
	0x000000,
	0x000000,
	0xffffff,	// white
	0x000000,
};
int color_num = 16;

int disable_gamma=0;
int disable_fade=0;

void loadRGB( Neopixel &neo, int shift, int scale)
{
	if( scale>255 ) scale=255;
	else if( scale<0 ) scale=0;
	
	shift %= color_num;
	if( shift < 0 ) shift += color_num;
	
	for(int i=0; i<neo.ByteNum()/3; i++,shift++){
		if( shift == color_num ) shift = 0;
		for(int c=0; c<3; c++){
			int col = (color[shift]>>(c*8))&0xff;
			if( !disable_fade )
				col = (col*scale)>>8;
			if( !disable_gamma )
				col = gamma_tbl[col];
			neo.rgb[3*i+c] = col;
		}
	}
	neo.rgb2grb();
}


enum COMMANDS {
	INVALID,HELP,NAMES,LEDS,COLS,PIN,SHIFT,GAMMA,FADE
};

char names[] = {"\n"
" 0 - GPIO17 (header #11), GPIO_GEN0\n"
" 1 - GPIO18 (header #12), GPIO_GEN1, PCM_CLK\n"
" 2 - GPIO21 (header #13), GPIO_GEN2, GPIO27 at Rev 2\n"
" 3 - GPIO22 (header #15), GPIO_GEN3\n"
" 4 - GPIO23 (header #16), GPIO_GEN4\n"
" 5 - GPIO24 (header #18), GPIO_GEN5\n"
" 6 - GPIO25 (header #22), GPIO_GEN6\n"
" 7 - GPIO04 (header #07), GPIO_GCLK\n"
" 8 - GPIO00 (header #03), SDA1 I2C, GPIO02 at Rev 2\n"
" 9 - GPIO01 (header #05), SCL1 I2C, GPIO03 at Rev 2\n"
"10 - GPIO08 (header #24), SPI0_CE0_N\n"
"11 - GPIO07 (header #26), SPI0_CE1_N\n"
"12 - GPIO10 (header #19), SPI0_MOSI\n"
"13 - GPIO09 (header #21), SPI0_MOSO\n"
"14 - GPIO11 (header #23), SPI0_CLK\n"
"15 - GPIO14 (header #08), UART0_TDX\n"
"16 - GPIO15 (header #10), UART0_RDX\n"
"17 - GPIO28 (header #24), trough GPIO08\n"
"18 - GPIO29 (header #21), trough GPIO09\n"
"19 - GPIO30 (header #19), trough GPIO10\n"
"20 - GPIO31 (header #23), trough GPIO11\n"
"21 - GPIO05 (header #29)\n"
"22 - GPIO06 (header #31)\n"
"23 - GPIO13 (header #33)\n"
"24 - GPIO19 (header #35)\n"
"25 - GPIO26 (header #37)\n"
"26 - GPIO12 (header #32)\n"
"27 - GPIO16 (header #36)\n"
"28 - GPIO20 (header #38)\n"
"29 - GPIO21 (header #40)\n"};

struct COMMAND_LIST
{
	const COMMANDS id;
	const char *command;
	const char *abbrev;
	const char *help;
	const int num_parameters;
}; 

COMMAND_LIST commands[] = {
	{HELP, "-help",		"?", "Print this message", 0},
	{NAMES,"-names",	"n", "Print pin name table", 0},
	{LEDS, "-led",		"l", "Number of leds <num>; default is 12",1},
	{COLS, "-colors",	"c", "Set a color  series <hexa,...>; max 100",1},
	{PIN,  "-pin",		"p", "I/O pin; default is 1",1},
	{SHIFT,"-time",		"t", "Time between shift in ms; default is 100; 0 is once",1},
	{GAMMA,"-gamma",	"g", "Disable gamma",0},
	{FADE, "-fade",		"f", "Disable fade",0}
};
const int num_commands = sizeof(commands)/sizeof(COMMAND_LIST);

void display_help(const char *app_name)
{
   fprintf(stderr, "\nControlling neopixel RGB strip\n");
   fprintf(stderr, "usage: %s [options]\n", app_name);
   fprintf(stderr, "Image parameter commands\n");
   for (int i = 0; i < num_commands; i++)
      fprintf(stderr, "-%s, -%s\t: %s\n", commands[i].abbrev, commands[i].command, commands[i].help);
   fprintf(stderr, "\n");
}

int get_params(int argc, const char **argv)
{	
	if( argc && argv ) {
		for (int i = 1; i < argc; i++) {
			COMMANDS command = INVALID;
			int num_parameters=0;
			const char *arg = argv[i];
			if (!arg)  continue;
			if (arg[0] == '-')  {
				for (int j = 0; j < num_commands; j++) {
					if (!strcmp(arg+1, commands[j].command) || !strcmp(arg+1, commands[j].abbrev)) {
						num_parameters = commands[j].num_parameters;
						command = commands[j].id;
						break;
					}
				}
			}
			else {
				fprintf(stderr,"Invalid \"%s\" parameter.\n",arg);
				display_help(basename(argv[0])); 
				return 1;
			}
			if( command == INVALID ) {
				fprintf(stderr,"Invalid \"%s\" option.\n",arg);
				display_help(basename(argv[0])); 
				return 1;
			}
			if(command != INVALID && num_parameters > 0 && (i+1 >= argc) ) {
				fprintf(stderr,"Missing parameter for \"%s\" option. Ignore option.\n",arg);
				continue;
			}
 			switch(command){
				case HELP:
					display_help(basename(argv[0])); 
					return 1;
				case GAMMA:
					disable_gamma = 1;
					break;
				case FADE:
					disable_fade = 1;
					break;
				case NAMES:
					printf(names); 
					return 1;
				case LEDS:
					sscanf(argv[i+1], "%u", &led_num);
					i++; // extra parameter
 					break;
				case COLS:
					{
						color_num=0;
						char c=',';
						for(char*ptr=(char*)argv[i+1]; c==',' && color_num<100; color_num++){
							color[color_num]=0;
							for(c=*ptr++; (c>='0'&&c<='9') || (c>='A'&&c<='F') || (c>='a'&&c<='f'); c=*ptr++){
								color[color_num] <<= 4;
								if(c>='a') c-='a'-10;
								else if(c>='A') c-='A'-10;
								else c -='0';
								color[color_num] |= c;
							}
							if( c != 0 && c != ',' ) {
								fprintf(stderr,"Illegal characther in color setting, %s\n",argv[i+1]);
								display_help(basename(argv[0])); 
								return 1;
							}
							printf("%06x ",color[color_num]);
						}
						printf("\n");
					}
					i++; // extra parameter
					break;
				case PIN:
					sscanf(argv[i+1], "%u", &pin_num);
					i++; // extra parameter
					break;
				case SHIFT:
					sscanf(argv[i+1], "%u", &shift_time);
					i++; // extra parameter
 					break;
				default: break;
			}			
		} 
	}
	return 0;
}

 
int main (int argc, const char *argv[])
{
	if( get_params(argc,argv) ) return 0;
	Neopixel neopixel( led_num, pin_num );
	printf("\nneopixel start, %d led, any key terminates program\n",led_num);
	neopixel.clear_at_end = shift_time!=0; // turn on/off leds at end
	getcharNoWait *gch = new getcharNoWait();
	for( int shift = 0, scale = 400, up = 0; ; shift ++ ) {
		loadRGB( neopixel, shift, scale);
		unsigned int micro = neopixel.send();	
		printf( "neopixel cycle: %d, %d us, repeat: %d\r", shift, micro, neopixel.repeat);
		if( !shift_time ) break;
		if( getchar()!=EOF ) break;
		delay(shift_time);
		if( up ) { scale+=7; if(scale>=400) scale = 400,up=0; }
		else { scale-=7;  if(scale<=17) scale = 17,up=1;}
		shift ++;
	}
	delete gch;
	
	printf( "\nneopixel ready\n\n");
	return 0 ;
}
